#Exercise 3 template for CHE210D

#import python modules
import numpy as np
import time
import math

#import compiled fortran library
import BHJ_lib

import sys
import os

import atomwrite

#whether or not to use the 3d visualization
UseVisual = True
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 2.5
#set the random number seed; useful for debugging
#np.random.seed(342324)


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.

def CalcPDF(Pos, Type, Hist):
    """Wrapper for PDF calculation.
Input:
    Pos: (N,3) position array
    Type: (N) type array
    Hist: (NBins) array of counts
Output:
    Hist: (NBins) array of counts
"""
    global L, NBins
    return BHJ_lib.calcpdf(Pos, Type, L, Hist)

def VVIntegrate(Pos, Type, Vel, Accel):
    """Wrapper for VVIntegrate.
Input:
    Pos: (N,3) position array
    Type: (N) type array
    Vel: (N,3) velocity array
    Accel: (N,3) Accel array
Output:
    PEnergy: potential energy, float
    KEnergy: kinetic energy, float
    Pos, Vel, Accel
"""
    global M, L, Cut, sigmaP, sigmaF, epsilonPP, epsilonFF, epsilonPF, massP, massF, dt
    return BHJ_lib.vvintegrate(Pos, Type, Vel, Accel, M, L, Cut, sigmaP, sigmaF, epsilonPP, epsilonFF, epsilonPF, massP, massF, dt)

def CalcEnergy(Pos, Type):
    """Wrapper for energy calculation.
Input:
    Pos: (N,3) position array
Output:
    PEnergy: potential energy, float
"""
    global M, L, Cut, sigmaP, sigmaF, epsilonPP, epsilonFF, epsilonPF
    return BHJ_lib.calcenergy(Pos, Type, M, L, Cut, sigmaP, sigmaF, epsilonPP, epsilonFF, epsilonPF)


def CalcEnergyForces(Pos, Type, Forces):
    """Wrapper for energy and force calculation.
Input:
    Pos: (N,3) position array
    Forces: (N,3) force array
Output:
    PEnergy: potential energy, float
    Forces: (N,3) force array (updated)
"""
    global M, L, Cut, sigmaP, sigmaF, epsilonPP, epsilonFF, epsilonPF, massP, massF
    return BHJ_lib.calcenergyforces(Pos, Type, M, L, Cut, sigmaP, sigmaF, epsilonPP, epsilonFF, epsilonPF, massP, massF, Forces)


def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = Dir / np.sqrt(np.sum(Dir * Dir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [CalcEnergy(Pos + NormDir * x, Type) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = CalcEnergy(Pos + NormDir * dx, Type)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [CalcEnergy(Pos + NormDir * 2. * dx, Type)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [CalcEnergy(Pos + NormDir * Dists[-1], Type)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = CalcEnergy(Pos + NormDir * d3, Type)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = CalcEnergy(Pos + NormDir * d3, Type)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    PE, Forces = CalcEnergyForces(Pos, Type, np.zeros_like(Pos))
    Dir = Forces
    OldPE = 1.e300
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = CalcEnergyForces(Pos, Type, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(NTot, M, NPoly, NFull, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    N: number of polymer chains
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((NTot,3), float)
    #make the type array (0=Polymer, 1=Fullerene)
    Type = np.zeros((NTot), int)
    #compute integer grid # of locations for cubic lattice
    NLat = int(NTot**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #make array to store which lattice sites posess polymer chains
    PolyPos = np.zeros((NLat, NLat, NLat), bool)
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    #Place all of the polymer chains then go back through and fill in fullerenes

    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= NTot:
                    return Pos, Type

                if i < NPoly*M:
                    Type[i] = 0
                else:
                    Type[i] = 1

    return Pos, Type


def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #find net momentum
    Mom = np.zeros(3, float)
    for i in range(np.shape(Vel)[0]):
        if Type[i] == 0:
            Mom += massP*Vel[i]
        elif Type[i] == 1:
            Mom += massF*Vel[i]
    #recenter to zero net momentum
    for i in range(np.shape(Vel)[0]):
        if Type[i] == 0:
            Vel[i] -= Mom/(massP*np.shape(Vel)[0])
        elif Type[i] == 1:
            Vel[i] -= Mom/(massF*np.shape(Vel)[0])

    #Scale velocities
    sum = 0
    for i in range(np.shape(Vel)[0]):
        if Type[i] == 0:
            sum += massP*np.sum(Vel[i]*Vel[i])
        elif Type[i] == 1:
            sum += massF*np.sum(Vel[i]*Vel[i])
    VScale = np.sqrt(T*len(Vel)/sum)
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #get the acceleration from the forces
    Accel = np.zeros_like(Pos)
    PEnergy, Forces = CalcEnergyForces(Pos, Type, np.zeros_like(Pos))
    for i in range(np.shape(Pos)[0]):
        if Type[i]==0:
            Accel[i] = Forces[i]/massP
        elif Type[i]==1:
            Accel[i] = Forces[i]/massF
    return Accel

###############
#####Main######
###############
NPAtom = 240. #static
M = 8
NPoly = NPAtom/M
Poly_Full_Ratio = 1
NFull = NPoly*(1/Poly_Full_Ratio)
NTot = NPAtom + NFull
Na = 6.02e23

#Tuneable parameters in the model
sigmaP = 1.
sigmaF = .75
epsilonPP = 1.
epsilonFF = .85
epsilonPF = .95
massP = 1.
massF = .75

#density in mg/cm^3
rho = 0.8
#number density = 0.8 atoms/V (L in meters)
L = ((NPAtom+NFull)/(rho))**(1./3.)

rc = Cut #defined golbally as Cut
tempK = 300.
kB = 1.38e-23
T = 1

#timestep should be in fs (e-15) range
dt = 0.0001

RescaleFreq = 50
PdbFreq = 10
EFreq = 100
HistFreq = 100
MSDFreq = 100
Dim = 3

#initialize histogram
NBins = 100
Hist = np.zeros((NBins), int)

#initialize pos and vel
Pos, Type = InitPositions(NTot, M, NPoly, NFull, L)

Names = []
for q in Type:
    if q == 0:
        Names.append('P')
    elif q == 1:
        Names.append('C')

PEnergy, Pos = ConjugateGradient(Pos, 0.001, 1.e-8, 1.e-10)
print "Conjugate Gradient Minumization Finished"
#write full pdb file
Pdb = atomwrite.pdbfile("M=" + str(M) + ".pdb", L, AtomNames = Names)
#write single file
#atomwrite.pdbfile("frame.pdb", L, AtomNames = Names).write(Pos)

Vel = InitVelocities(NTot, T)
Accel = InitAccel(Pos)

# 1ns
NStepsEquil1 = 10000
# 1ns
NStepsEquil2 = 10000
# 5ns
NStepsProd = 400000

statsfile = file("M = " + str(M) + " (stats).txt", "w")
statsfile.write("Temperature: " + str(tempK) + "\n")
statsfile.write("Simulation cube length (m): " + str(L) + "\n")
statsfile.write("Density (molecules/m^3): " + str(rho) + "\n")
statsfile.write("Degree of Polymerization: " + str(M) + "\n")
statsfile.write("Eq Time 1: " + str(dt*NStepsEquil1) + "\n")
statsfile.write("Eq Time 2: " + str(dt*NStepsEquil2) + "\n")
statsfile.write("Production Time: " + str(dt*NStepsProd) + "\n")
statsfile.write("Number of Polymer Chains: " + str(NPoly) + "\n")
statsfile.write("Number of Fullerenes: " + str(NFull) + "\n")
statsfile.write("Polymer/Full Ratio: " + str(NPoly/NFull) + "\n")
statsfile.write("sigmaF: " + str(sigmaF) + "\n")
statsfile.write("sigmaP: " + str(sigmaP) + "\n")
statsfile.write("sigmaF/sigmaP Ratio: " + str(sigmaF/sigmaP) + "\n")
statsfile.write("epsilonPP: " + str(epsilonPP) + "\n")
statsfile.write("epsilonFF: " + str(epsilonFF) + "\n")
statsfile.write("epsilonPF: " + str(epsilonPF) + "\n")
statsfile.write("mass monomer (kg/mol): " + str(massP) + "\n")
statsfile.write("mass fullerene (kg/mol): " + str(massF) + "\n")
statsfile.close()

energyfile = file("M = " + str(M) + " (energy).txt", "w")
energyfile.write("M=" + str(M) + "\n")
energyfile.write("time\tEnergy\n")

#Equilibration Period 1
time = 0
for i in range(0, NStepsEquil1-1):
    Pos, Vel, Accel, kenergy, penergy = VVIntegrate(Pos, Type, Vel, Accel)
    if i % EFreq == 0:
        print str(i) + "\t" + str(penergy) + "\t" + str(kenergy)
        energyfile.write(str(time) + "\t" + str(kenergy+penergy) + "\n")
    if i % RescaleFreq == 0:
        Vel = RescaleVelocities(Vel, T)
    time = time + dt
print "Equilibriation Step 1 Complete"

#Equilibration Period 2
EnergyValues = []
for j in range(0, NStepsEquil2-1):
    Pos, Vel, Accel, kenergy, penergy = VVIntegrate(Pos, Type, Vel, Accel)
    EnergyValues.append(kenergy + penergy)
    if j % EFreq == 0:
        print str(j) + "\t" + str(penergy) + "\t" + str(kenergy)
        energyfile.write(str(time) + "\t" + str(kenergy+penergy) + "\n")
    if j % RescaleFreq == 0:
        Vel = RescaleVelocities(Vel, T)
    time = time + dt
print "Equilibriation Step 2 Complete"

#Instantaneous rescaling
AvgEnergyForStep2 = np.mean(EnergyValues)
newKineticE = AvgEnergyForStep2-penergy
curKineticE = 0
for i in range(np.shape(Vel)[0]):
    if Type[i] == 0:
        curKineticE += 0.5*massP*np.sum(Vel[i]*Vel[i])
    elif Type[i] == 1:
        curKineticE += 0.5*massF*np.sum(Vel[i]*Vel[i])
ScaleFactor = np.sqrt(newKineticE/curKineticE)
print "Velocity Rescale Factor: " + str(ScaleFactor) + "\n"
Vel = ScaleFactor*Vel

Pos0 = Pos.copy()

outfile = file("M = " + str(M) + " (MSD).txt", "w")
outfile.write("M=" + str(M) + "\n")
outfile.write("time(fs)\tMSD(m^2)\n")

#Production Run
time = 0.
displacement2 = 0.
numObs = 0.
for k in range(0, NStepsProd-1):
    Pos, Vel, Accel, kenergy, penergy = VVIntegrate(Pos, Type, Vel, Accel)
    if k % EFreq == 0:
        print str(k) + "\t" + str(penergy) + "\t" + str(kenergy)
        energyfile.write(str((NStepsEquil1 + NStepsEquil2)*dt+ time) + "\t" + str(kenergy+penergy) + "\n")
    if k % HistFreq == 0:
        Hist = CalcPDF(Pos, Type, Hist)
        numObs +=1
    if k % PdbFreq == 0:
        Pdb.write(Pos)
    if k % MSDFreq == 0:
        displacement2 = np.sum((Pos - Pos0)**2)/NTot
        outfile.write(str(((NStepsEquil1 + NStepsEquil2)*dt+ time)) + "\t" + str(displacement2) + "\n")
    time = time + dt

#Output PDF information
PDFfile = file("M = " + str(M) + " (Hist).txt", "w")
PDFfile.write("M=" + str(M) + "\n")
PDFfile.write("r\tg(r)\n")
PDF = ((Hist/(numObs*((NFull*(NFull-1))/2)))*(L**3/(((4*math.pi*(L/NBins)**3)/3)*((NBins+1)**3-NBins**3))))
for (index,value) in enumerate(PDF):
    PDFfile.write(str((index*(L/NBins))/sigmaF) + "\t" + str(value) + "\n")
PDFfile.close()

energyfile.close()
outfile.close()
Pdb.close()

print "Simulation Complete"